""" SQI definitions and some internal calculations based on:

H. Tang, M. Wang, Y. Hu, B. Guo, and T. Li, “Automated Signal Quality Assessment for Heart Sound Signal by Novel Features and Evaluation in Open Public Datasets,” BioMed Research International, vol. 2021. Hindawi Limited, pp. 1–15, Feb. 24, 2021.

https://www.hindawi.com/journals/bmri/2021/7565398/
"""

import numpy as np
import math
import scipy.stats as scistat
import scipy.fft as fft
import pyPCG.preprocessing as preproc
from scipy.linalg import hankel
from scipy.spatial.distance import pdist
from pyPCG import pcg_signal

def _sample_entropy_fast(x, m, r):
    # from Tang et al.
    x = x-np.mean(x)
    x = x/np.std(x)
    N = len(x)
    indm = hankel(np.arange(N-m), np.arange(N-m,N-1))
    inda = hankel(np.arange(N-m), np.arange(N-m,N))
    ym = x[indm]
    ya = x[inda]
    cheb = pdist(ym, "chebychev")
    cm = np.sum(cheb<=r)*2/ (ym.shape[0]*(ym.shape[0]-1))
    cheb = pdist(ya, "chebychev")
    ca = np.sum(cheb<=r)*2/ (ya.shape[0]*(ya.shape[0]-1))
    return -math.log(ca/cm)

def _autocorr(x):
    result = np.correlate(x, x, mode='full')
    result = result/np.max(result)
    return result[result.size//2+1:]

def _nextpow2(x):
    return math.ceil(math.log(x, 2))

def _fast_cfsd(sig,f1,f2,k):
    # from Tang et al.
    w = np.exp(-1j*2*np.pi*(f2-f1)/(k*sig.fs))
    a = np.exp(1j*2*np.pi*f1/sig.fs)
    x = preproc.envelope(sig).data
    x = x-np.mean(x)
    m = len(x)
    nfft = 2**_nextpow2(m+k-1)
    kk = np.arange(-m,max(k,m))
    kk2 = (kk**2)/2
    ww = w**kk2
    nn = np.arange(m)
    aa = a**(-nn)
    aa = aa*ww[m+nn]
    y = x * aa
    fy = fft.fft(y,nfft)
    fv = fft.fft(1/ww[:k-1+m],nfft)
    fy = fy * fv #type: ignore
    g = fft.ifft(fy)
    g = g[m:m+k-1] * ww[m:m+k-1]
    return np.abs(g)

def periodicity_score(sig: pcg_signal, f1:float=0.3, f2:float=2.5, k:int=200) -> float:
    """Calculate SQI based on cylcic periodicity.
    
    Larger -> Better

    Args:
        sig (pcg_signal): input signal (envelope)
        f1 (float, optional): min Hz to search. Defaults to 0.3.
        f2 (float, optional): max Hz to search. Defaults to 2.5.
        k (int, optional): weight parameter. Defaults to 200.

    Returns:
        float: Periodicity score
    """
    gamma = _fast_cfsd(sig,f1,f2,k)
    return np.max(gamma)/np.median(gamma)

def sentropy(sig: pcg_signal, win:int=2, r:float=0.2, precalc:bool=False) -> float:
    """Calculate sample entropy SQI
    
    Smaller -> Better

    Args:
        sig (pcg_signal): input signal
        win (int, optional): window size. Defaults to 2.
        r (float, optional): weight parameter. Defaults to 0.2.
        precalc (bool, optional): is the input a precalculated envelope. Defaults to False.

    Returns:
        float: sample entropy
    """
    temp = preproc.envelope(sig) if not precalc else sig
    env = preproc.resample(temp,30)
    return _sample_entropy_fast(env.data,win,r)

def autocorr_max(sig: pcg_signal, bpm_min: float=100, bpm_max: float=200) -> float:
    """Calculate SQI based on autocorrelation.
    
    Larger -> Better

    Args:
        sig (pcg_signal): input signal (envelope)
        bpm_min (float, optional): minimum expected BPM. Defaults to 100.
        bpm_max (float, optional): maximum expected BPM. Defaults to 200.

    Returns:
        float: autocorrelation maximum
    """
    ar = _autocorr(sig.data)
    expect_min = round((1/bpm_max)*sig.fs)
    expect_max = round((1/bpm_min)*sig.fs)
    m = np.max(ar[expect_min:expect_max])
    return m

def env_std(sig: pcg_signal, precalc: bool=False) -> float:
    """Calculate standard deviation of the envelope values
    
    Smaller -> Better

    Args:
        sig (pcg_signal): input signal
        precalc (bool, optional): is the input a precalculated envelope. Defaults to False.

    Returns:
        float: standard deviation of envelope
    """
    env = preproc.envelope(sig) if not precalc else sig
    s = np.std(env.data).astype(float)
    return s

def raw_kurt(sig: pcg_signal) -> float:
    """Calculate kurtosis of raw signal values
    
    Larger -> Better

    Args:
        sig (pcg_signal): input signal

    Returns:
        float: kurtosis of signal
    """
    return scistat.kurtosis(sig.data) #type: ignore


import numpy as np
import pandas as pd
import numpy.typing as npt
import scipy.stats as sts
from scipy.special import erfcinv
from typing import Callable, TypedDict

def trim_transform(data: npt.NDArray[np.float64], trim_precent: float) -> npt.NDArray[np.float64]:
    """Trim the upper and lower percentage of values

    Args:
        data (np.ndarray): input data to trim
        trim_precent (float): percentage to trim away

    Returns:
        np.ndarray: trimmed values
    """
    return sts.trimboth(data,trim_precent/100)

def outlier_remove_transform(data: npt.NDArray[np.float64], dist: float=3.0) -> npt.NDArray[np.float64]:
    """Remove outliers based on the MAD (median of absolute differences)

    Args:
        data (np.ndarray): input data
        dist (float, optional): MAD score threshold. Defaults to 3.0.

    Returns:
        np.ndarray: data without outliers
    """
    d = np.abs(data - np.median(data))
    c = -1/(np.sqrt(2)*erfcinv(3/2))
    mdev = c*np.median(d)
    s = d/mdev if mdev else 0.
    return data[s<dist]

def mean(data: npt.NDArray[np.float64]) -> npt.NDArray[np.float64] | np.float64:
    """Calculate mean of inputs

    Args:
        data (np.ndarray): input data (can be 2 dimensional array)

    Returns:
        np.ndarray | float: mean value of data, if input is 2D then return the value along of 1st axis
    """
    if len(data.shape) == 1:
        return np.mean(data)
    else: return np.mean(data,axis=1)

def std(data: npt.NDArray[np.float64]) -> npt.NDArray[np.float64] | np.float64:
    """Calculate standard deviation of inputs

    Args:
        data (np.ndarray): input data (can be 2 dimensional array)

    Returns:
        np.ndarray | float: standard deviation of data, if input is 2D then return the value along of 1st axis
    """
    if len(data.shape) == 1:
        return np.std(data)
    else: return np.std(data,axis=1)

def rms(data: npt.NDArray[np.float64]) -> npt.NDArray[np.float64] | np.float64:
    """Calculate root mean square of inputs

    Args:
        data (np.ndarray): input data (can be 2 dimensional array)

    Returns:
        np.ndarray | float: root mean square of data, if input is 2D then return the value along of 1st axis
    """
    if len(data.shape) == 1:
        return np.sqrt(np.mean(data**2))
    else: return np.sqrt(np.mean(data**2,axis=1))

def med(data: npt.NDArray[np.float64]) -> npt.NDArray[np.float64] | np.float64:
    """Calculate median of inputs

    Args:
        data (np.ndarray): input data (can be 2 dimensional array)

    Returns:
        np.ndarray | float: median of data, if input is 2D then return the value along of 1st axis
    """
    if len(data.shape) == 1:
        return np.median(data)
    else: return np.median(data,axis=1)

def skew(data: npt.NDArray[np.float64]) -> npt.NDArray[np.float64] | np.float64:
    """Calculate skewness of inputs

    Args:
        data (np.ndarray): input data (can be 2 dimensional array)

    Returns:
        np.ndarray | float: skewness of data, if input is 2D then return the value along of 1st axis
    """
    if len(data.shape) == 1:
        return sts.skew(data)
    else: return sts.skew(data,axis=1)

def kurt(data: npt.NDArray[np.float64]) -> npt.NDArray[np.float64] | np.float64:
    """Calculate kurtosis of inputs

    Args:
        data (np.ndarray): input data (can be 2 dimensional array)

    Returns:
        np.ndarray | float: kurtosis of data, if input is 2D then return the value along of 1st axis
    """
    if len(data.shape) == 1:
        return sts.kurtosis(data) #type: ignore
    else: return sts.kurtosis(data,axis=1) #type: ignore

def max(data: npt.NDArray[np.float64],k: int=1) -> npt.NDArray[np.float64] | np.float64:
    """Get maximum values from input

    Args:
        data (np.ndarray): input data
        k (int, optional): number of largest values to return. Defaults to 1.

    Returns:
        np.ndarray | float: maximum value(s)
    """
    s_data = np.sort(data)
    select = s_data[-k-1:-1]
    ret = select[::-1]
    if len(ret)==1:
        ret=ret[0]
    return ret

def min(data: npt.NDArray[np.float64],k: int=1) -> npt.NDArray[np.float64] | np.float64:
    """Get minimum values from input

    Args:
        data (np.ndarray): input data
        k (int, optional): number of smallest values to return. Defaults to 1.

    Returns:
        np.ndarray | float: minimum value(s)
    """
    s_data = np.sort(data)
    ret = s_data[0:k]
    if len(ret)==1:
        ret=ret[0]
    return ret

def percentile(data: npt.NDArray[np.float64], perc: float=25) -> npt.NDArray[np.float64] | np.float64:
    """Calculate given percentile of inputs

    Args:
        data (np.ndarray): input data (can be 2 dimensional array)
        perc (float): selected percentile to calculate. Defaults to 25.

    Returns:
        np.ndarray | float: given percentile of data, if input is 2D then return the value along of 1st axis
    """
    if len(data.shape) == 1:
        return np.percentile(data,perc)
    else: return np.percentile(data,perc,axis=1)

def window_operator(data: npt.NDArray[np.float64],win_size: int,fun: Callable,overlap_percent: float=0.5) -> tuple[npt.NDArray[np.int_],npt.NDArray[np.float64]]:
    """Apply given statistical function over a sliding window on the input

    Args:
        data (np.ndarray): input data
        win_size (int): window size
        fun (Callable): statistical function to apply
        overlap_percent (float, optional): window overlap as a ratio to the window size. Defaults to 0.5.

    Returns:
        tuple[np.ndarray,np.ndarray]: window sample locations (usually used as time dimension), and calculated values in the windows
    """
    step = win_size-round(win_size*overlap_percent)
    val,loc = [], []
    for i in range(0,len(data) - win_size,step):
        val.append(fun(data[i:i+win_size]))
        loc.append(i)
    return np.array(loc), np.array(val)

def iqr(data: npt.NDArray[np.float64]) -> npt.NDArray[np.float64] | np.float64:
    """Calculate interquartile range

    Args:
        data (np.ndarray): input data

    Returns:
        np.ndarray | float: interquartile range of data
    """
    return percentile(data,75)-percentile(data,25)

class stats_config(TypedDict):
    """Type to hold statistic calculation configs"""
    calc_fun: Callable
    """function for calculation"""
    name: str
    """name of the calculated statistic"""

class stats_group:
    """Group statistic calculations together for reuse
    
    Attributes:
        configs (list[stats_config]): List of statistic calculations with names
        signal_stats (dict[str, dict[str, list[float]]]): Signal statistics by segment, #TODO: this will become its own type in the future
        dataframe (pd.DataFrame): Pandas dataframe container of statistics for utility
        
    Example:
        Create a statistic group calculation:
        
        For an easier experience use the `stats_config` type
        
        >>> import pyPCG.stats as sts
        >>> stat_1 = {"calc_fun":sts.mean,"name":"Mean"}
        >>> stat_2 = {"calc_fun":sts.std,"name":"Std"}
        >>> mean_std = sts.stats_group(stat_1,stat_2)
        
        Run the created group with some dummy features:
        
        >>> import numpy as np
        >>> dummy = {"length":np.arange(10),"max freq":np.arange(10)}
        >>> basic_stats = mean_std.run(dummy)
        >>> print(basic_stats)
        {'Feature': ['length', 'max freq'], 'Mean': [4.5, 4.5], 'Std': [2.8722813232690143, 2.8722813232690143]}
        
        Add statistics for a given segment, and export it as xlsx:
        
        >>> mean_std.add_stat("Test",basic_stats)
        >>> mean_std.export("test.xlsx")
        
    """
    def __init__(self,*stats: stats_config) -> None:
        self.configs = []
        self.signal_stats = {}
        self.dataframe = pd.DataFrame()
        for stat in stats:
            self.configs.append(stat)

    def _update_df(self):
        first_key = next(iter(self.signal_stats.keys()))
        ftr_and_stats = list(self.signal_stats[first_key].keys())
        temp = {"Segment":[]}
        for key in ftr_and_stats:
            temp[key] = []
        for segment,dat in self.signal_stats.items():
            for key,vals in dat.items():
                for val in vals:
                    temp[key].append(val)
            for _ in dat["Feature"]:
                temp["Segment"].append(segment)
        self.dataframe = pd.DataFrame(temp)

    def run(self,ftr_dict: dict[str,npt.NDArray[np.float64]]) -> dict[str,list[float]]:
        """Run the statistic calculation based on configuration

        Args:
            ftr_dict (dict[str,np.ndarray]): feature dictionary, same format as the output of feature_group.run

        Returns:
            dict[str,list[float]]: Calculated statistics, named
        """
        ret = {"Feature":[]}
        for stat in self.configs:
            ret[stat["name"]] = []
        for name,value in ftr_dict.items():
            ret["Feature"].append(name)
            for stat in self.configs:
                ret[stat["name"]].append(stat["calc_fun"](value))
        return ret
    
    def add_stat(self,segment:str, stats:dict[str,list[float]]):
        """Add calculated statistics to signal_stats with the given segment name

        Args:
            segment (str): segment name to save to
            stats (dict[str,list[float]]): calculated statistics
        """
        self.signal_stats[segment] = stats
        self._update_df()
    
    def calc_group_stats(self,total_ftr_dict: dict[str,dict[str,npt.NDArray[np.float64]]]):
        """Calculate all stats on all given features and segments

        Args:
            total_ftr_dict (dict[str,dict[str,np.ndarray]]): Feature dictionaries named by segment
        """
        for segment,ftr_dict in total_ftr_dict.items():
            self.add_stat(segment,self.run(ftr_dict))

    def export(self,filename: str):
        """Export statistics to excel file

        Args:
            filename (str): filename to save to
        """
        with pd.ExcelWriter(filename) as writer:
            self.dataframe.to_excel(excel_writer=writer,sheet_name="Summary",index=False)
            for segment in self.dataframe["Segment"].unique():
                sub = self.dataframe[self.dataframe["Segment"]==segment][self.dataframe.columns.difference(["Segment"],sort=False)]
                sub.to_excel(excel_writer=writer,sheet_name=segment,index=False)

def calc_group_stats(ftr_dict: dict[str,dict[str,npt.NDArray[np.float64]]], *configs: tuple[Callable,str]) -> dict[str,list[float]]:
    """Calculate the same statistics for different segments and their features

    Args:
        ftr_dict (dict[str,dict[str,np.ndarray]]): input segment features, <segment name>:<feature dict from featuregroup.run>

    Returns:
        dict[str,list[float]]: statistics for each segment and its features
    """
    cols = {"Segment":[],"Feature":[]}
    for config in configs:
        cols[config[1]] = []
    for seg, ftrs in ftr_dict.items():
        for ftr,val in ftrs.items():
            cols["Segment"].append(seg)
            cols["Feature"].append(ftr)
            for config in configs:
                cols[config[1]].append(config[0](val))
    return cols

def export_stats(filename: str, group_stats: dict[str,list[float]]):
    """Export statistics calculated with calc_group_stats to excel

    Args:
        filename (str): name of excel file
        group_stats (dict[str,list[float]]): input statistics
    """
    df = pd.DataFrame(group_stats)
    with pd.ExcelWriter(filename) as writer:
        df.to_excel(excel_writer=writer,sheet_name="Summary",index=False)
        for segment in df["Segment"].unique():
            sub = df[df["Segment"]==segment][df.columns.difference(["Segment"],sort=False)]
            sub.to_excel(excel_writer=writer,sheet_name=segment,index=False)

if __name__ == '__main__':
    print("Statistics")


import csv
import warnings
import numpy as np
import numpy.typing as npt
import scipy.io as sio
from math import floor

def read_signal_file(path: str, format: str) -> tuple[npt.NDArray[np.int_],int]:
    """Read in fetal heartsound containing file
    
    Supported file formats:
    
    - `wav`: wav file
    - `mat`: MATLAB file (containing two variables: `fs`-samplerate, `sig`-signal data)
    - `raw`: raw binary (headerless)
    - `FETA`: every second byte is PCG data (headerless)
    - `1k`: 1 kB chunks (CURRENTLY UNTESTED)

    Args:
        path (str): Path to input file
        format (str): File format identification

    Returns:
        tuple[np.ndarray,int]: Unprocessed heartsound signal read in from file, and sample rate in Hz. If the input file was headerless, then the value is 0
    
    Example:
        Read in a 1 min wav file with 333 Hz samplerate
        
        >>> import pyPCG.io as pcg_io
        >>> data, fs = pcg_io.read_signal_file("example.wav","wav")
        >>> print(len(data),fs)
        19980 333
        
        Raw binary file reading:
        >>> import pyPCG.io as pcg_io
        >>> data, fs = pcg_io.read_signal_file("example.dat","raw")
        >>> print(len(data),fs)
        19980 0
    """
    
    signal = np.array([])
    fs = 0

    with open(path, 'rb') as dat:
        data = np.array(list(dat.read()))
        if format == 'raw':
            signal = data
        elif format == '1k':
            warnings.warn("1k format is not yet tested. Use at your own risk")
            BLOCK_SIZE = 1024
            block_count = floor(len(data)/BLOCK_SIZE)
            for i in range(block_count):
                signal = np.append(signal,data[25+i*BLOCK_SIZE:1024+i*BLOCK_SIZE])
            if block_count*BLOCK_SIZE+24 < len(data):
                signal = np.append(signal,data[25+block_count*BLOCK_SIZE:])
        elif format == '2nd':
            
            def _get_byte_offset_start(b_sig):
                test = b_sig[:16]
                mark_00 = 0
                for i in range(8):
                    if sum(test[i::8][:2])==0:
                        mark_00 = i
                        break
                test_ofs = mark_00
                read_ofs = mark_00+1%2
                return test_ofs, read_ofs

            def _find_byte_shift(b_sig,offset=0):
                test = b_sig[offset::8]
                mask = test!=0
                loc = np.where(mask)[0]
                if len(loc)>0:
                    res = ((loc[0])*8+offset)
                    # print(f"Byte shift detected at:{res}")
                    return res
                return -1

            def _correct_byte_shift(b_sig,t_offset=0):
                has_shift = True
                while has_shift:
                    shift = _find_byte_shift(b_sig,t_offset)
                    if shift<0:
                        has_shift = False
                    else:
                        b_sig = np.append(b_sig[:shift-1],b_sig[shift:])
                return b_sig
            
            t_offset,r_offset = _get_byte_offset_start(data)
            corr = _correct_byte_shift(data,t_offset)
            signal = corr[r_offset::2]
        elif format == 'wav':
            fs, signal = sio.wavfile.read(path)
        elif format == 'mat':
            mat = sio.loadmat(path)
            fs = mat["fs"][0,0]
            signal = np.squeeze(mat["sig"])
        else:
            raise ValueError('Format not recognized')
    return np.array(signal), fs

def read_hsannot_file(fpath: str) -> tuple[list[float],list[float]]:
    """Reads manually labeled heartsounds from annotation csv file
    
    The csv format has to be the following:
    
    - `Location`: heartsound location in seconds
    - `Value`: heartsound type of `"S1"` or `"S2"`

    Args:
        fpath (str): path to annotation file

    Returns:
        tuple[list[float],list[float]]: S1 and S2 annotation locations
    """
    
    s1_loc, s2_loc = [],[]
    with open(fpath,'r') as annot:
        reader = csv.DictReader(annot,delimiter=';')
        for line in reader:
            if line['Value']=='S1':
                s1_loc.append(float(line['Location']))
            elif line['Value']=='S2': 
                s2_loc.append(float(line['Location']))
            else:
                print("Unknown label. Skipping...")
    return s1_loc,s2_loc

if __name__ == '__main__':
    print("Data Loader")


from enum import Enum
import numpy as np
import numpy.typing as npt
import scipy.signal as sgnl
from pyPCG import pcg_signal
import pyPCG.lr_hsmm as lr_hsmm

def adv_peak(signal: pcg_signal, percent_th:float=0.5) -> tuple[npt.NDArray[np.float64],npt.NDArray[np.int_]]:
    """Adaptive peak detection, based on local maxima and following value drop

    Args:
        signal (pcg_signal): input signal to detect peaks (usually this is the envelope)
        percent_th (float, optional): percent drop in value to be considered a real peak. Defaults to 0.5.

    Returns:
        tuple[np.ndarray,np.ndarray]: detected peak values and their locations
    """
    peaks = []
    sig = signal.data
    loc_max,_ = sgnl.find_peaks(sig)
    for loc_ind in range(len(loc_max)):
        search_start = loc_max[loc_ind]
        search_end = loc_max[loc_ind+1] if loc_ind+1 < len(loc_max) else len(sig)
        th_val = sig[search_start]*percent_th
        if np.any(sig[search_start:search_end]<th_val):
            peaks.append(search_start)
    return np.array(sig[peaks]), np.array(peaks)

def peak_sort_diff(peak_locs: npt.NDArray[np.int_]) -> tuple[npt.NDArray[np.int_],npt.NDArray[np.int_]]:
    """Sort detected peaks based on time differences.
    
    A short time difference corresponds with systole -> S1-S2, a long time difference corresponds with diastole -> S2-S1

    Args:
        peak_locs (np.ndarray): Detected peak locations in samples

    Raises:
        ValueError: Less than two peaks detected

    Returns:
        tuple[np.ndarray,np.ndarray]: S1 and S2 locations
    """
    if len(peak_locs)<2:
        raise ValueError("Too few peak locations (<2)")

    interval_diff = np.append(np.diff(peak_locs,2),[0,0]) # type: ignore
    s1_loc = peak_locs[interval_diff>0]
    s2_loc = peak_locs[interval_diff<0]
    return s1_loc, s2_loc

def segment_peaks(peak_locs: npt.NDArray[np.int_], envelope_signal: pcg_signal ,start_drop:float=0.6, end_drop:float=0.6) -> tuple[npt.NDArray[np.int_],npt.NDArray[np.int_]]:
    """Create start and end locations from the detected peaks based on the provided envelope.
    
    The relative drop in envelope value is marked as the start and end positions of the given heartsound

    Args:
        peak_locs (np.ndarray): detected peak locations in samples
        envelope_signal (pcg_signal): precalculated envelope of the signal (homomorphic recommended)
        start_drop (float, optional): precent drop in value for start location. Defaults to 0.6.
        end_drop (float, optional): percent drop in value for end location. Defaults to 0.6.

    Returns:
        tuple[np.ndarray,np.ndarray]: heartsound boundary locations
    """
    envelope = envelope_signal.data
    starts, ends = [],[]
    for peak_ind,peak_loc in enumerate(peak_locs):
        prev_peak = peak_locs[peak_ind-1] if peak_ind>0 else 0
        next_peak = peak_locs[peak_ind+1] if peak_ind+1<len(peak_locs) else len(envelope)
        start_th = envelope[peak_loc]*start_drop
        end_th = envelope[peak_loc]*end_drop
        prev_drop = np.nonzero(envelope[prev_peak:peak_loc]<start_th)[0]
        next_drop = np.nonzero(envelope[peak_loc:next_peak]<end_th)[0]
        if len(prev_drop)!=0 and len(next_drop)!=0:
            starts.append(prev_drop[-1]+prev_peak)
            ends.append(next_drop[0]+peak_loc)
    return np.array(starts), np.array(ends)

def load_hsmm(path:str) -> lr_hsmm.LR_HSMM:
    """Load pretrained LR-HSMM model.
    
    Note:
        Training is done internally, it is not recommended to use it right now

    Args:
        path (str): path to pretrained model json file

    Returns:
        lr_hsmm.LR_HSMM: pretrained model loaded in
    """
    model = lr_hsmm.LR_HSMM()
    model.load_model(path)
    return model

def segment_hsmm(model:lr_hsmm.LR_HSMM,signal:pcg_signal) -> npt.NDArray[np.float64]:
    """Use a trained LR-HSMM model to segment a pcg signal

    Args:
        model (lr_hsmm.LR_HSMM): trained LR-HSMM model
        signal (pcg_signal): input signal to be segmented

    Raises:
        ValueError: Samplerate discrepancy

    Returns:
        np.ndarray: heartcycle states
    """
    if(model.signal_fs!=signal.fs):
        raise ValueError(f"Unexpected signal samplerate {signal.fs}, LR-HSMM expects {model.signal_fs}")
    states, _ = model.segment_single(signal.data)
    return states

class heart_state(Enum):
    """Heart states enum"""
    S1 = 1
    SYS = 2
    S2 = 3
    DIA = 4
    unknown = 0
heart_state.S1.__doc__ = "First heartsound"
heart_state.S2.__doc__ = "Second heartsound"
heart_state.SYS.__doc__ = "Systole section"
heart_state.DIA.__doc__ = "Diastole section"
heart_state.unknown.__doc__ = "Default value, unknown state"

def convert_hsmm_states(states: npt.NDArray[np.float64], state_id: int|heart_state) -> tuple[npt.NDArray[np.int_],npt.NDArray[np.int_]]:
    """Convert selected LR-HSMM state to start and end times

    Args:
        states (np.ndarray): output states of LR-HSMM
        state_id (int, heart_state): selected state to convert

    Raises:
        ValueError: Unrecognized heart cycle state

    Returns:
        tuple[np.ndarray,np.ndarray]: state boundaries in samples
    """
    if (type(state_id) is int) and (state_id not in [0,1,2,3,4]):
        raise ValueError(f"Unrecognized heart cycle state: {state_id}")
    if type(state_id) is heart_state:
        state_id = int(state_id.value)
    select = np.zeros_like(states)
    select[states==state_id] = 1
    states_diff = np.diff(select)
    state_start = np.nonzero(states_diff>0)[0]
    state_end = np.nonzero(states_diff<0)[0]
    return state_start, state_end

if __name__ == '__main__':
    print("Segmenting and peak detection")


"""
Logistic Regression - Hidden Semi-Markov Model segmenter

Based on:
D. B. Springer, L. Tarassenko and G. D. Clifford, "Logistic Regression-HSMM-Based Heart Sound Segmentation," in IEEE Transactions on Biomedical Engineering, vol. 63, no. 4, pp. 822-832, April 2016, doi: 10.1109/TBME.2015.2475278.
"""

import pywt
import json
import warnings
import numpy as np
import numpy.typing as npt
import scipy.signal as sgn
from math import ceil
from sklearn.linear_model import LogisticRegression
from hsmmlearn.emissions import AbstractEmissions
from hsmmlearn.hsmm import HSMMModel
from scipy.stats import norm

INF = 999999999

class LR_HSMM():
    """
    Main segmenter object

    Attributes:
        signal_fs (int): Sampling frequency of the input signal. Default: 1000 [Hz]
        feature_fs (int): Sampling frequency for feature calculations. Default: 50 [Hz]
        mean_s1_len (float): Average S1 duration. Default: 122 [ms]
        mean_s2_len (float): Average S2 duration. Default: 99 [ms]
        std_s1_len (float): Standard deviation of S1 durations. Default: 22 [ms]
        std_s2_len (float): Standard deviation of S2 durations. Default: 22 [ms]
        bandpass_frq (tuple[float,float]): Cutoff frequencies for pre-processing band-pass filtering (Butterworth, 4th order). Default: (25,400) [Hz]
        expected_hr_range (tuple[float,float]): Minimum and maximum expected heartrates. Default: (30,120) [bpm]

        hsmm_model (hsmmlearn.hsmm.HSMMModel): State predictor model
        lr_model (hsmmlearn.emissions.AbstractEmissions): Probability emissions for the HSMM states. Includes a LogisticRegression model for each state
    """

    def __init__(self) -> None:
        self.signal_fs = 1000
        self.feature_fs = 50
        self.mean_s1_len = 122
        self.mean_s2_len = 99
        self.std_s1_len = 22
        self.std_s2_len = 22
        self.bandpass_frq = (25,400)
        self.expected_hr_range = (30,120)

        self.hsmm_model = None
        self.lr_model = _LREmission()

    def train_model(self,train_data:npt.NDArray[np.float64],train_s1_annot:(npt.NDArray[np.float64]|npt.NDArray[np.int_]),train_s2_annot:npt.NDArray[np.float64]) -> None:
        """Trains the model on the specified data with S1 and S2 location annotations 

        Args:
            train_data (np.ndarray): Array of input signals
            train_s1_annot (np.ndarray): Array of S1 annotations for each signal
            train_s2_annot (np.ndarray): Array of S2 annotations for each signal
        """

        tmat = np.array([[0.,1.,0.,0.],[0.,0.,1.,0.],[0.,0.,0.,1.],[1.,0.,0.,0.]])
        states = np.array([])
        f_henv, f_env, f_psd, f_wt = np.array([]),np.array([]),np.array([]),np.array([])
        d_hr, d_sys = np.array([]),np.array([])
        print("Generating features...")
        for i,(data,s1_annot,s2_annot) in enumerate(zip(train_data,train_s1_annot,train_s2_annot)):
            hr, sys = _get_hr_sys(data,self.signal_fs,self.bandpass_frq,self.expected_hr_range[0],self.expected_hr_range[1])
            sts = _generate_states(data,s1_annot,s2_annot,self.signal_fs,self.feature_fs,self.mean_s1_len,self.mean_s2_len,self.std_s1_len,self.std_s2_len)
            henv, env, psd, wt = _generate_features(data,self.signal_fs,self.feature_fs,self.bandpass_frq)
            states = np.append(states,sts)
            f_henv = np.append(f_henv,henv)
            f_env = np.append(f_env,env)
            f_psd = np.append(f_psd,psd)
            f_wt = np.append(f_wt,wt)
            d_hr = np.append(d_hr,hr)
            d_sys = np.append(d_sys,sys)
        features = np.array([f_henv,f_env,f_psd,f_wt])
        durs = _get_duration_distributions(np.mean(d_hr),np.mean(d_sys),self.feature_fs,self.mean_s1_len,self.mean_s2_len,self.std_s1_len,self.std_s2_len)
        print("Training model...")
        self.lr_model = _LREmission(features.T, states)
        self.hsmm_model = HSMMModel(self.lr_model,durs,tmat)

    def segment_single(self,sig:npt.NDArray[np.float64]) -> tuple[npt.NDArray[np.float64],npt.NDArray[np.float64]]:
        """Predicts the states for the given PCG signal

        Args:
            sig (np.ndarray): Input signal to be segmented

        Returns:
            tuple[np.ndarray,np.ndarray]: Predicted state for each sample and the calculated features
        """

        henv, env, psd, wt = _generate_features(sig,self.signal_fs,self.feature_fs,self.bandpass_frq)
        seg_features = np.array([henv, env, psd, wt], dtype=np.float64)
        if self.hsmm_model is None:
            warnings.warn("Attempting to segment with untrained model. Returning empty states...",RuntimeWarning)
            return np.empty(0), np.empty(0)
        d_states = self.hsmm_model.decode(seg_features.T)
        e_states = _expand_states(d_states+1,self.feature_fs,self.signal_fs,len(sig))
        return e_states, seg_features

    def save_model(self,filename:str) -> None:
        """Saves the model parameters to a json file

        Args:
            filename (str): Name of file to be saved
        """
        if self.hsmm_model is None:
            warnings.warn("Attempting to save untrained model. No file will be written. Returning...",RuntimeWarning)
            return
        durs = self.hsmm_model.durations.tolist() #type: ignore
        tmat = self.hsmm_model.tmat.tolist() #type: ignore
        emissions = self.hsmm_model.emissions.serialize() #type: ignore
        config = {
            "sig_fs":self.signal_fs,
            "f_fs": self.feature_fs,
            "preproc_bp": self.bandpass_frq,
            "mean_s1": self.mean_s1_len,
            "mean_s2" : self.mean_s2_len,
            "std_s1" : self.std_s1_len,
            "std_s2" : self.std_s2_len,
            "min_hr": self.expected_hr_range[0],
            "max_hr": self.expected_hr_range[1]
        }
        serialized_hsmm = {"config":config,"durations":durs,"transition":tmat,"emissions":emissions}
        with open(filename,"w") as save:
            save.write(json.dumps(serialized_hsmm))

    def load_model(self,filename:str) -> None:
        """Loads the model parameters from a json file

        Args:
            filename (str): Name of file containing model parameters
        """

        with open(filename,"r") as load:
            data = json.loads(load.read())
            durs = np.array(data["durations"])
            tmat = np.array(data["transition"])
            config = data["config"]

            self.signal_fs = config["sig_fs"]
            self.feature_fs = config["f_fs"]
            self.bandpass_frq = config["preproc_bp"]
            self.mean_s1_len = config["mean_s1"]
            self.mean_s2_len = config["mean_s2"]
            self.std_s1_len = config["std_s1"]
            self.std_s2_len = config["std_s2"]
            self.expected_hr_range = (config["min_hr"],config["max_hr"])

            emission = _LREmission()
            emission.unserialize(data["emissions"])
            self.lr_model = emission
            self.hsmm_model = HSMMModel(self.lr_model,durs,tmat)

class _LREmission(AbstractEmissions):

    def __init__(self, features=None, states=None) -> None:
        self.LRmodel_s1 = LogisticRegression()
        self.LRmodel_s2 = LogisticRegression()
        self.LRmodel_sys = LogisticRegression()
        self.LRmodel_dia = LogisticRegression()
        self.predictors = [self.LRmodel_s1,self.LRmodel_sys,self.LRmodel_s2,self.LRmodel_dia]
        if(features is None or states is None):
            return

        s1_train, s2_train, sys_train, dia_train = np.zeros_like(states), np.zeros_like(states), np.zeros_like(states), np.zeros_like(states)
        s1_train[states != 1] = 1
        s2_train[states != 3] = 1
        sys_train[states != 2] = 1
        dia_train[states != 4] = 1

        self.LRmodel_s1 = LogisticRegression(random_state=0,max_iter=1000,class_weight="balanced").fit(features,s1_train)
        self.LRmodel_s2 = LogisticRegression(random_state=0,max_iter=1000,class_weight="balanced").fit(features,s2_train)
        self.LRmodel_sys = LogisticRegression(random_state=0,max_iter=1000,class_weight="balanced").fit(features,sys_train)
        self.LRmodel_dia = LogisticRegression(random_state=0,max_iter=1000,class_weight="balanced").fit(features,dia_train)
        self.predictors = [self.LRmodel_s1,self.LRmodel_sys,self.LRmodel_s2,self.LRmodel_dia]

    def likelihood(self, obs):
        l_s1 = self.LRmodel_s1.predict_proba(obs)[:,0]
        l_s2 = self.LRmodel_s2.predict_proba(obs)[:,0]
        l_sys = self.LRmodel_sys.predict_proba(obs)[:,0]
        l_dia = self.LRmodel_dia.predict_proba(obs)[:,0]
        return np.array([l_s1,l_sys,l_s2,l_dia])

    def serialize(self):
        serialized_models = {"lr_s1":None, "lr_sys":None, "lr_s2":None, "lr_dia":None}
        for lr_type,model in zip(serialized_models.keys(),self.predictors):
            params = model.get_params()
            attrs = [i for i in dir(model) if i.endswith('_') and not i.endswith('__') and not i.startswith('_')]
            attr_dict = {i: getattr(model, i) for i in attrs}
            for k in attr_dict:
                if isinstance(attr_dict[k], np.ndarray):
                    attr_dict[k] = attr_dict[k].tolist()
            serialized_lr = {"params":params,"attrs":attr_dict}
            serialized_models[lr_type] = serialized_lr #type: ignore
        return serialized_models
    
    def unserialize(self,serial):
        for loaded,lr in zip(serial.keys(),self.predictors):
            params = serial[loaded]["params"]
            attrs = serial[loaded]["attrs"]
            lr.set_params(**params)
            for k in attrs:
                if isinstance(attrs[k],list):
                    setattr(lr,k,np.array(attrs[k]))
                else:
                    setattr(lr,k,attrs[k])

def _envelope_feature(sig):
    env = abs(sgn.hilbert(sig)) #type: ignore
    return env

def _h_envelope_feature(sig,sig_fs):
    env = _envelope_feature(sig)
    lp = sgn.butter(6,8,output='sos',fs=sig_fs,btype='lowpass')
    filt = np.exp(sgn.sosfiltfilt(lp,np.log(env)))
    filt[0] = filt[1] #?
    return filt

def _psd_feature(sig,sig_fs):
    f_lo = 40
    f_hi = 60
    [f,_,Zxx] = sgn.stft(sig,fs=sig_fs,window="hamming",nperseg=sig_fs//40,scaling="psd",nfft=1024)
    lo_pos = np.argmin(np.abs(f-f_lo))
    hi_pos = np.argmin(np.abs(f-f_hi))
    psd = np.mean(np.abs(Zxx[lo_pos:hi_pos,:]),axis=0)
    psd_re = sgn.resample_poly(psd,len(sig),len(psd))
    return psd_re

def _wt_feature(sig):
    coefs = pywt.wavedec(sig,"rbio3.9",level=3)
    cD = coefs[1:]
    cD.reverse()
    detail = np.zeros((3,len(sig)))
    for i in range(3):
        d = np.tile(cD[i],(2**(i+1),1)).T.ravel()
        start = len(d)-len(sig)-1//2
        end = start + len(sig)
        detail[i,:] = d[start:end]

    d3 = np.abs(detail[2,:])
    return d3

def _normalize(sig):
    m = np.mean(sig)
    s = np.std(sig)
    n_sig = (sig-m)/s
    return n_sig

def _spike_removal(sig,sig_fs):
    window_s = round(sig_fs/2)
    trailing = len(sig) % window_s
    frames = np.reshape(sig[:-trailing],(window_s,-1))
    MAAs = np.max(np.abs(frames),axis=0)
    while(np.count_nonzero(MAAs>np.median(MAAs)*3)>0):
        # val = np.max(MAAs,axis=1)
        framenum = np.argmax(MAAs,axis=0)
        if type(framenum) != np.int64:
            if len(framenum)>1:
                framenum = framenum[1]
        pos = np.argmax(np.abs(frames[:,framenum]))
        zerocross = np.where(np.abs(np.diff(np.sign(frames[:,framenum])))>0)[0]
        last = np.nonzero(zerocross[0:pos])[0][-1]
        start = max(1,last)
        zerocross[0:pos] = 0
        end = min(np.nonzero(zerocross)[0][0],window_s)
        frames[start:end,framenum] = 0.0001
        MAAs = np.max(np.abs(frames),axis=0)
    removed = np.reshape(frames,(-1,1))
    removed = np.append(removed,sig[len(removed):])
    return removed

def _generate_features(sig,sig_fs,f_fs,preproc=(25,400)):
    bpf = sgn.butter(4,preproc,"bandpass",output="sos",fs=sig_fs)
    f_sig = sgn.sosfiltfilt(bpf,sig)
    # rem_sig = _spike_removal(f_sig,sig_fs)
    rem_sig = f_sig

    h_env = _normalize(_h_envelope_feature(rem_sig,sig_fs))
    env = _normalize(_envelope_feature(rem_sig))
    psd = _normalize(_psd_feature(rem_sig,sig_fs))
    wt = _normalize(_wt_feature(rem_sig))

    d_h_env = sgn.resample_poly(h_env,sig_fs,f_fs)
    d_env = sgn.resample_poly(env,sig_fs,f_fs)
    d_psd = sgn.resample_poly(psd,sig_fs,f_fs)
    d_wt = sgn.resample_poly(wt,sig_fs,f_fs)

    return d_h_env,d_env,d_psd,d_wt

def _generate_states(sig,annot_s1,annot_s2,sig_fs,f_fs,mean_s1=122,mean_s2=99,std_s1=22,std_s2=22):
    henv = _h_envelope_feature(sig,sig_fs)
    env = sgn.resample_poly(henv,sig_fs,f_fs)
    states = np.zeros_like(env)
    ms_scale = sig_fs/1000
    fs_scale = sig_fs/f_fs
    scale = f_fs/1000
    as1 = np.round(np.array(annot_s1)*scale).astype(int) # *ms_scale/fs_scale
    as2 = np.round(np.array(annot_s2)*scale).astype(int)
    ms1 = round(mean_s1*scale)
    ms2 = round(mean_s2*scale)
    ss1 = round(std_s1*scale)
    ss2 = round(std_s2*scale)
    for s1 in as1:
        upper_s1 = min(len(states)-1,s1+ms1+ss1) # +std_s1
        lower_s1 = max(1,s1-ms1-ss1)
        search = env[lower_s1:upper_s1]
        s1_ind = np.argmax(search)
        s1_ind = min(len(states)-1,lower_s1+s1_ind) #type: ignore
        upper_s1 = min(len(states)-1,ceil(s1_ind+(ms1/2)))
        lower_s1 = max(0,ceil(s1_ind-(ms1/2)))

        states[lower_s1:upper_s1] = 1
    for s2 in as2:
        upper_s2 = min(len(states)-1,s2+ms2+ss2)
        lower_s2 = max(0,s2-ms2-ss2)
        search = env[lower_s2:upper_s2]*(1-states[lower_s2:upper_s2])
        s2_ind = np.argmax(search)
        s2_ind = min(len(states)-1,lower_s2+s2_ind) #type: ignore

        upper_s2 = min(len(states)-1,ceil(s2_ind+(ms2/2)))
        lower_s2 = max(0,ceil(s2_ind-(ms2/2)))
        states[lower_s2:upper_s2] = 3

        s1_labels = np.nonzero(states == 1)[0]
        diffs = s1_labels - s2
        diffs[diffs<0] = INF
        end_pos = 0
        if len(diffs<INF)==0:
            end_pos = len(states)-1
        else:
            end_pos = s1_labels[np.argmin(diffs)]
        states[ceil(s2_ind+(ms2/2)):end_pos] = states[ceil(s2_ind+(ms2/2)):end_pos]+4 #type: ignore

    first_definite = np.nonzero(states)[0][0]
    if first_definite > 0:
        if states[first_definite+1] == 1:
            states[0:first_definite] = 4
        if states[first_definite+1] == 3:
            states[0:first_definite] = 2
    last_definite = np.nonzero(np.array(states))[0][-1]
    if last_definite > 0:
        if states[last_definite] == 1:
            states[last_definite:] = 2
        if states[last_definite] == 3:
            states[last_definite:] = 4

    states[states==0] = 2
    return states

def _get_hr_sys(sig,sig_fs,preproc=(25,400),min_hr=30,max_hr=120):
    bpf = sgn.butter(4,preproc,"bandpass",output="sos",fs=sig_fs)
    f_sig = sgn.sosfiltfilt(bpf,sig)
    # rem_sig = spike_removal(f_sig,sig_fs)
    h_env = _h_envelope_feature(f_sig,sig_fs)
    y = h_env - np.mean(h_env)
    coef = sgn.correlate(y,y)
    coef = coef[len(h_env):]/np.max(coef)
    min_ind = round((60/max_hr)*sig_fs)
    max_ind = round((60/min_hr)*sig_fs)
    ind = np.argmax(coef[min_ind:max_ind]) + min_ind
    heartrate = 60/(ind/sig_fs)
    
    max_sys = round(((60/heartrate)*sig_fs)/2)
    min_sys = round(0.2*sig_fs)
    ind = np.argmax(coef[min_sys:max_sys]) + min_sys
    systole = ind/sig_fs
    # print(systole)
    # systole = 0.1 # temporary hack
    return heartrate, systole

def _get_duration_params(hr,sys,mean_s1=122,mean_s2=99,std_s1=22,std_s2=22):
    mean_sys = round(sys*1000)
    std_sys = 25
    # mean_dia = round((60/hr)-sys)*1000-94
    mean_dia = round(((60/hr)-sys)*1000)-130
    # std_dia = round(0.07*mean_dia) + 6
    std_dia = round(0.15*mean_dia)

    min_sys = mean_sys - 3*(std_sys+std_s1)
    max_sys = mean_sys + 3*(std_sys+std_s1)
    min_dia = mean_dia - 3*std_dia
    max_dia = mean_dia + 3*std_dia
    min_s1 = mean_s1 - 3*std_s1
    max_s1 = mean_s1 + 3*std_s1
    min_s2 = mean_s2 - 3*std_s2
    max_s2 = mean_s2 + 3*std_s2

    # max_duration = (60/hr)*1000
    max_duration = max([max_s1+2*std_s1,max_s2+2*std_s2,max_sys+2*(std_sys+std_s1),max_dia+2*std_dia])
    # min_duration = min([min_s1-2*std_s1,min_s2-2*std_s2,min_sys-2*(std_sys-std_s1),min_dia-2*std_dia])
    return max_duration, mean_sys, std_sys, mean_dia, std_dia

def _get_duration_distributions(hr,sys,f_fs,mean_s1=122,mean_s2=99,std_s1=22,std_s2=22):
    max_duration, mean_sys, std_sys, mean_dia, std_dia = _get_duration_params(hr,sys,mean_s1,mean_s2,std_s1,std_s2)
    # fs_scale = round(sig_fs/f_fs)
    time_scale = f_fs/1000
    max_duration = round(max_duration*time_scale)
    s1_dur, s2_dur, sys_dur, dia_dur = np.zeros((max_duration)),np.zeros((max_duration)),np.zeros((max_duration)),np.zeros((max_duration))
    for i in range(max_duration):
        s1_dur[i] = norm.pdf(i,loc=mean_s1*time_scale,scale=std_s1*time_scale)
        s2_dur[i] = norm.pdf(i,loc=mean_s2*time_scale,scale=std_s2*time_scale)
        sys_dur[i] = norm.pdf(i,loc=mean_sys*time_scale,scale=std_sys*time_scale)
        dia_dur[i] = norm.pdf(i,loc=mean_dia*time_scale,scale=std_dia*time_scale)
    return np.array([s1_dur,sys_dur,s2_dur,dia_dur])

def _diff(sig):
    sub = np.insert(sig,0,0)
    sig = sig-sub[:-1]
    return sig[1:]

def _expand_states(states,orig_fs,new_fs,new_len):
    expanded = np.zeros(new_len)
    changes = np.nonzero(np.diff(states))[0]
    changes = np.append(changes,len(states)-1)
    start = 0
    for end in changes:
        mid = round((end-start)/2) + start
        mid_val = states[mid]
        exp_start = round((start/orig_fs)*new_fs)
        exp_end = round((end/orig_fs)*new_fs) if round((end/orig_fs)*new_fs) < new_len else new_len
        expanded[exp_start:exp_end] = mid_val
        start = end
    return expanded

if __name__ == '__main__':
    print("LR-HSMM model based on Springer et al.")


import numpy as np
import numpy.typing as npt
import matplotlib.pyplot as plt
import copy

__version__ = "0.1-a"

class pcg_signal:
    """PCG signal object. This is used to store the signal data, as well as its sampling rate. The object also includes a processing log to track the processing steps and settings done to the signal.
    
    When using the plot function the last step is used as a default figure title.
    
    All processing steps require an object like this as input, and their output is also a signal object.
    
    Attributes:
        data (np.ndarray): signal data
        fs (int): sampling rate in Hz
        processing_log (list[str]): processing steps and parameters on the signal
    
    Example:
        Creating a signal from numpy array:
        
        >>> import pyPCG
        >>> import numpy as np 
        >>> sig = np.arange(10)
        >>> signal = pyPCG.pcg_signal(sig,100,["Example signal"])
        >>> print(signal)
        PCG signal [0.1s 100Hz] ['Example signal']
        
        Using `pyPCG.io.read_signal_file()`:
        
        >>> import pyPCG
        >>> import pyPCG.io
        >>> # reading in a 60 second long wav file with 333 Hz sampling rate
        >>> sig, fs = pyPCG.io.read_signal_file("example.wav","wav")
        >>> signal = pyPCG.pcg_signal(sig,fs)
        >>> # or optionally using the splat operator
        >>> signal = pyPCG.pcg_signal(*pyPCG.io.read_signal_file("example.wav","wav"))
        >>> print(signal)
        PCG signal [60s 333Hz] ['File read in']
    """
    def __init__(self, data:npt.NDArray[np.int_|np.float64], fs:int=1, log:list[str]|None=None) -> None:
        """Create PCG signal object

        Args:
            data (np.ndarray): raw signal data
            fs (int, optional): sampling rate. Defaults to 1.
            log (list[str] | None, optional): processing log. Defaults to None.

        Raises:
            ValueError: Tried to create PCG signal with no data
        """
        if len(data) == 0:
            raise ValueError("Tried to create PCG signal with no data")
        self.data = data.astype(float)
        self.fs = fs
        self.processing_log = ["File read in"] if log is None else log
        
    def __repr__(self) -> str:
        return f"PCG signal [{self.get_timelength()}s {self.fs}Hz] {self.processing_log}"

    def get_timelength(self) -> float:
        """Get length of signal in seconds

        Returns:
            float: length of signal in seconds
        """
        return len(self.data)/self.fs

def zero_center(sig: pcg_signal) -> pcg_signal:
    """Center signal to zero

    Args:
        sig (pcg_signal): Input signal

    Returns:
        pcg_signal: Centered signal
    """
    ret_sig = copy.deepcopy(sig)
    ret_sig.data -= np.mean(ret_sig.data)
    ret_sig.processing_log.append("Zero center")
    return ret_sig

def unit_scale(sig: pcg_signal) -> pcg_signal:
    """Scale signal to [-1,1] interval

    Args:
        sig (pcg_signal): Input signal

    Returns:
        pcg_signal: Scaled signal
    """
    ret_sig = copy.deepcopy(sig)
    ret_sig.data /=np.max(np.abs(ret_sig.data))
    ret_sig.processing_log.append("Unit scale")
    return ret_sig

def std_scale(sig: pcg_signal) -> pcg_signal:
    """Scale signal to 1 std

    Args:
        sig (pcg_signal): Input signal

    Returns:
        pcg_signal: Scaled signal
    """
    ret_sig = copy.deepcopy(sig)
    ret_sig.data /=np.std(ret_sig.data)
    ret_sig.processing_log.append("Std scale")
    return ret_sig

def normalize(sig: pcg_signal) -> pcg_signal:
    """Center to zero and scale signal to [-1,1] interval

    Args:
        sig (pcg_signal): Input signal

    Returns:
        pcg_signal: Normalized signal
    """
    return unit_scale(zero_center(sig))

def plot(sig: pcg_signal, zeroline: bool=False, xlim: tuple|None=None) -> None:
    """Plot pcg signal with appropriate dimensions

    Args:
        sig (pcg_signal): signal to plot
        zeroline (bool, optional): plot a dashed line at zero. Defaults to False.
        xlim (tuple, optional): set horizontal limits on the plot. Where time is the horizontal axis and it is measured in seconds.
    """
    t = sig.get_timelength()
    plt.plot(np.linspace(0,t,len(sig.data)),sig.data)
    if zeroline:
        z_line = np.zeros_like(sig.data)
        plt.plot(np.linspace(0,t,len(sig.data)),z_line,"k:")
    plt.title(sig.processing_log[-1])
    plt.xlabel("Time (s)")
    plt.ylabel("PCG (a.u.)")
    if xlim is not None:
        plt.xlim(xlim)

def multiplot(*args):
    for sig in args:
        time = np.linspace(0,sig.get_timelength,len(sig.data))
        plt.plot(time,sig.data)

if __name__ == '__main__':
    print("Signal container and process pipeline builder")


from setuptools import setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="pyPCG_toolbox",
    version="0.1-b",
    description="A PCG processing toolbox",
    author="Kristóf Müller, Janka Hatvani, Miklós Koller, Márton Áron Goda",
    author_email="muller.kristof@itk.ppke.hu",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mulkr/pyPCG-toolbox/",
    project_urls={"Bug Tracker": "https://github.com/mulkr/pyPCG-toolbox/issues",},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    packages=["pyPCG"],
    install_requires=[  "scikit-learn>=1.1.3",
                        "scipy>=1.9.3",
                        "PyWavelets>=1.5.0",
                        "pandas>=2.0.0",
                        "matplotlib>=3.6.2",
                        "numpy>=1.25.2",
                        "hsmmlearn @ git+https://github.com/jvkersch/hsmmlearn@master",
                        "emd>=0.6.2",
                        "nolds>=0.5.2",
                        "tqdm>=4.64.1",
                        "pyQt5>=5.15.7"],
    python_requires=">=3.10"
)


import warnings
import nolds
import numpy as np
import numpy.typing as npt
import scipy.ndimage as ndimage
import scipy.fft as fft
import pywt
from typing import Callable, Literal, TypedDict
from pyPCG import pcg_signal

def _check_start_end(start,end):
    if len(start) != len(end):
        warnings.warn("Start and end arrays not the same size. Converting to same size...")
        if len(start) > len(end):
            start = start[:len(end)]
        else:
            end = end[:len(start)]
    return start, end

def time_delta(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_], sig: pcg_signal) -> npt.NDArray[np.float64]:
    """Calculate time differences between pairs of points

    Args:
        start (np.ndarray): start points in samples
        end (np.ndarray): end points in samples
        sig (pcg_signal): input signal

    Returns:
        np.ndarray: time differences in seconds
    """
    start, end = _check_start_end(start,end)
    return (end-start)/sig.fs

def ramp_time(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],envelope: pcg_signal,type: str="onset") -> npt.NDArray[np.float64]:
    """Calculate ramp time (onset or exit), the time difference between the boundary and peak

    Args:
        start (np.ndarray): start points in samples
        end (np.ndarray): end points in samples
        envelope (pcg_signal): precalculated envelope signal
        type (str, optional): ramp type "onset" or "exit". Defaults to "onset".

    Raises:
        ValueError: ramp type not "onset" or "exit"

    Returns:
        np.ndarray: ramp times in seconds
    """
    start, end = _check_start_end(start,end)
    ret = []
    for s,e in zip(start,end):
        peak = np.argmax(envelope.data[s:e])
        l = e-s
        if type=="onset":
            ret.append(peak)
        elif type=="exit":
            ret.append(l-peak)
        else:
            raise ValueError("Unrecognized ramp type")
    return np.array(ret)/envelope.fs

def zero_cross_rate(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],sig: pcg_signal) -> npt.NDArray[np.float64]:
    """Calculate zero cross rate

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        sig (pcg_signal): input signal

    Returns:
        np.ndarray: zero cross rates
    """
    start, end = _check_start_end(start,end)
    ret = []
    for s, e in zip(start,end):
        cross_log = np.abs(np.diff(np.sign(sig.data[s:e])))>1
        cross_ind = np.nonzero(cross_log)[0]
        crosses = len(cross_ind)+0.5
        ret.append(crosses/(e-s))
    return np.array(ret)

def peak_spread(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],envelope: pcg_signal,factor: float=0.7) -> npt.NDArray[np.int_]:
    """Calculate peak spread, the amount of area under the peak with a given percentage of the total and time differences between the beginning and end

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        envelope (pcg_signal): input envelope signal
        factor (float, optional): percentage of total area. Defaults to 0.7.

    Returns:
        np.ndarray: time difference between the beginning and end of the percentage area in ms
    """
    start, end = _check_start_end(start,end)
    ret = []
    for s, e in zip(start, end):
        win = envelope.data[s:e]
        th = np.sum(win)*factor
        grow_left = True
        offset_left, offset_right = 0, 0
        peak = np.argmax(win)
        for _ in win:
            if np.sum(win[peak-offset_left:peak+offset_right])>=th:
                break
            if grow_left and peak-offset_left>0:
                offset_left += 1
                grow_left = not grow_left
                continue
            elif not grow_left and peak+offset_right<len(win)-2:
                offset_right += 1
                grow_left = not grow_left
                continue
        spread = offset_right-offset_left
        ret.append(spread/envelope.fs * 1000)
    return np.array(ret)

def peak_centroid(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],envelope: pcg_signal) -> tuple[npt.NDArray[np.int_],npt.NDArray[np.float64]]:
    """Calculate centroid (center of mass) of the envelope

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        envelope (pcg_signal): input envelope signal

    Returns:
        np.ndarray: time delays from start to centroid
        np.ndarray: envelope values at centroid
    """
    start, end = _check_start_end(start,end)
    loc, val = [], []
    for s, e in zip(start,end):
        win = envelope.data[s:e]
        th = np.sum(win)*0.5 #type: ignore
        centr = np.nonzero(np.cumsum(win)>th)[0][0]
        loc.append(centr)
        val.append(win[centr])
    return np.array(loc), np.array(val)

def peak_width(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],envelope: pcg_signal,factor: float=0.7) -> npt.NDArray[np.int_]:
    """Calculate conventional width of the given peak, the difference between the preceding and succeeding time locations where the value is at a given proportion of the peak value

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        envelope (pcg_signal): input envelope signal
        factor (float, optional): proportionality factor. Defaults to 0.7.

    Returns:
        np.ndarray: peak width in ms
    """
    start, end = _check_start_end(start,end)
    ret = []
    for s,e in zip(start,end):
        win = envelope.data[s:e]
        loc = np.argmax(win)
        th = win[loc]*factor
        w_prev = np.nonzero(win[:loc]<th)[0]
        w_next = np.nonzero(win[loc:]<th)[0]
        w_s = w_prev[-1] if len(w_prev)!=0 else 0
        w_e = w_next[0]+loc if len(w_next)!=0 else e-s
        width = w_e-w_s
        ret.append(width/envelope.fs * 1000)
    return np.array(ret)

def max_freq(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],sig: pcg_signal,nfft: int=512) -> tuple[npt.NDArray[np.float64],npt.NDArray[np.float64]]:
    """Calculate frequency with maximum amplitude of the segment

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        sig (pcg_signal): input signal
        nfft (int, optional): fft width parameter. Defaults to 512.

    Returns:
        np.ndarray: frequencies of the maximum amplitude
        np.ndarray: values of the maximum amplitude frequency
    """
    start, end = _check_start_end(start,end)
    freqs = np.linspace(0,sig.fs//2,nfft//2)
    loc, val = [],[]
    for s,e in zip(start,end):
        spect = abs(fft.fft(sig.data[s:e],n=nfft)) #type: ignore
        spect = spect[:nfft//2]
        loc.append(freqs[np.argmax(spect)])
        val.append(np.max(spect))
    return np.array(loc), np.array(val)

def spectral_spread(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],sig: pcg_signal, factor: float=0.7, nfft: int=512) -> npt.NDArray[np.int_]:
    """Calculate spectral spread of the segments, percentage of the total power of the segment and the frequency difference between the beginning and end of the calculated area
    
    For a more detailed definition, see `peak_spread`.

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        sig (pcg_signal): input signal
        factor (float, optional): percentage of total power. Defaults to 0.7.
        nfft (int, optional): fft width parameter. Defaults to 512.

    Returns:
        np.ndarray: difference of the beginning and end of the given area
    """
    start, end = _check_start_end(start,end)
    ret = []
    for s, e in zip(start, end):
        spect = abs(fft.fft(sig.data[s:e],n=nfft)) #type: ignore
        spect = spect[:nfft//2]
        power = spect**2
        th = np.sum(power)*factor
        grow_left = True
        offset_left, offset_right = 0, 0
        peak = np.argmax(power)
        for _ in power:
            if np.sum(power[peak-offset_left:peak+offset_right])>=th:
                break
            if grow_left and peak-offset_left>0:
                offset_left += 1
                grow_left = not grow_left
                continue
            elif not grow_left and peak+offset_right<len(power)-2:
                offset_right += 1
                grow_left = not grow_left
                continue
        spread = offset_right-offset_left
        ret.append(spread)
    return np.array(ret)

def spectral_centroid(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],sig: pcg_signal,nfft: int=512) -> tuple[npt.NDArray[np.float64],npt.NDArray[np.float64]]:
    """Calculate spectral centroid (center of mass)
    
    For a more detailed definition, see `peak_centroid`.

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        sig (pcg_signal): input signal
        nfft (int, optional): fft width parameter. Defaults to 512.

    Returns:
        np.ndarray: spectral centroid locations in Hz
        np.ndarray: spectral centroid values
    """
    start, end = _check_start_end(start,end)
    freqs = np.linspace(0,sig.fs//2,nfft//2)
    loc, val = [],[]
    for s,e in zip(start,end):
        spect = abs(fft.fft(sig.data[s:e],n=nfft)) #type: ignore
        spect = spect[:nfft//2]
        th = np.sum(spect)*0.5
        idx = np.nonzero(np.cumsum(spect)>th)[0][0]
        loc.append(freqs[idx])
        val.append(spect[idx])
    return np.array(loc), np.array(val)

def spectral_width(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],sig: pcg_signal, factor: float=0.7, nfft: int=512) -> npt.NDArray[np.int_]:
    """Calculate conventional width of the spectrum, the difference between the preceding and succeeding frequency locations where the value is at a given proportion of the maximum value
    
    For a more detailed definition, see `peak_width`.

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        sig (pcg_signal): input signal
        factor (float, optional): proportionality factor. Defaults to 0.7.
        nfft (int, optional): fft width parameter. Defaults to 512

    Returns:
        np.ndarray: spectral width
    """
    start, end = _check_start_end(start,end)
    ret = []
    for s, e in zip(start, end):
        spect = abs(fft.fft(sig.data[s:e],n=nfft)) #type: ignore
        spect = spect[:nfft//2]
        power = spect**2
        loc = np.argmax(power)
        val = power[loc]
        th = val*factor
        w_prev = np.nonzero(power[:loc]<th)[0]
        w_next = np.nonzero(power[loc:]<th)[0]
        w_s = w_prev[-1] if len(w_prev)!=0 else 0
        w_e = w_next[0]+loc if len(w_next)!=0 else len(power)
        ret.append(w_e-w_s)
    return np.array(ret)

def spectrum_raw(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],sig: pcg_signal,nfft:int=512) -> npt.NDArray[np.float64]:
    """Calculate spectra of all input segments

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        sig (pcg_signal): input signal
        nfft (int, optional): fft width parameter. Defaults to 512.

    Returns:
        np.ndarray: spectrum of each segment (2D)
    """
    start, end = _check_start_end(start,end)
    ret = []
    for s,e in zip(start,end):
        spect = abs(fft.fft(sig.data[s:e],n=nfft)) #type: ignore
        spect = spect[:nfft//2]
        ret.append(spect)
    return np.array(ret)

def max_cwt(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],sig: pcg_signal) -> tuple[npt.NDArray[np.float64],npt.NDArray[np.float64]]:
    """Calculate the maximum cwt coeffitient location in both directions

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        sig (pcg_signal): input signal

    Returns:
        np.ndarray: maximum locations in time
        np.ndarray: maximum locations in frequency
    """
    warnings.warn("CWT calculation done with PyWT which has parity problems with Matlab")
    start, end = _check_start_end(start,end)
    time,freq = [],[]
    for s,e in zip(start,end):
        coef, fr = pywt.cwt(sig.data[s:e],np.arange(1,100),"cmor1.0-1.5",sampling_period=1/sig.fs)
        coef = np.abs(coef)
        loc = np.unravel_index(np.argmax(coef),coef.shape)
        time.append(loc[0])
        freq.append(fr[loc[1]])
    return np.array(time),np.array(freq)

def cwt_peakdist(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],sig: pcg_signal, size:int=8) -> npt.NDArray[np.float64]:
    """Calculate the distance of the largest two peaks in the continuous wavelet transform of the segment.
    
    Note:
        This currently has parity problems with Matlab's CWT function

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        sig (pcg_signal): input signal
        size (int, optional): size of maximum filter, i.e. peak neighborhood to check local maximum. Defaults to 8.

    Returns:
        np.ndarray: distance of two largest peaks in CWT, if only one peak is detected the value is 0
    """
    warnings.warn("CWT calculation done with PyWT, which has parity problems with Matlab")
    start, end = _check_start_end(start,end)
    ret = []
    for s,e in zip(start,end):
        coef, fr = pywt.cwt(sig.data[s:e],np.arange(1,40),"cmor1.0-1.5",sampling_period=1/sig.fs)
        coef = np.abs(coef)
        ind = ndimage.maximum_filter(coef,size)==coef
        val = coef[ind]
        asc = np.argsort(val)[::-1]
        x,y = np.nonzero(ind)
        max_x = x[asc[:2]]
        max_y = y[asc[:2]]
        if len(max_x)<2 or len(max_y)<2:
            ret.append(0)
        else:
            ret.append(np.sqrt((max_x[0]-max_x[1])**2+(max_y[0]-max_y[1])**2))
    return np.array(ret)

def dwt_intensity(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],sig: pcg_signal,wt_family:str="db6",decomp_level:int=4,select_level:int=2) -> npt.NDArray[np.float64]:
    """Calculate the 'intensity' of discrete wavelet transform

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        sig (pcg_signal): input signal
        wt_family (str, optional): wavelet family to use for decomposition. Defaults to "db6".
        decomp_level (int, optional): wavelet decomposition level. Defaults to 4.
        select_level (int, optional): selected decomposition level. Defaults to 2.

    Returns:
        np.ndarray: intensities of selected wavelet detail
    """
    start, end = _check_start_end(start,end)
    ret = []
    decomp = pywt.wavedec(sig.data,wt_family,level=decomp_level)
    select = decomp[-select_level]
    scale = 2**select_level
    for s,e in zip(start,end):
        win = select[s//scale:e//scale]
        intens = np.sqrt(np.mean(win**2))
        ret.append(intens)
    return np.array(ret)

def dwt_entropy(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],sig: pcg_signal,wt_family:str="db6",decomp_level:int=4,select_level:int=2)  -> npt.NDArray[np.float64]:
    """Calculate Shannon entropy of selected discrete wavelet decomposition level

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        sig (pcg_signal): input signal
        wt_family (str, optional): wavelet family to use for decomposition. Defaults to "db6".
        decomp_level (int, optional): wavelet decomposition level. Defaults to 4.
        select_level (int, optional): selected decomposition level. Defaults to 2.

    Returns:
        np.ndarray: entropies of selected wavelet detail
    """
    start, end = _check_start_end(start,end)
    ret = []
    decomp = pywt.wavedec(sig.data,wt_family,level=decomp_level) #wavedec first, then window?
    select = decomp[-select_level]
    scale = 2**select_level
    for s,e in zip(start,end):
        win = select[s//scale:e//scale]
        ent = -np.sum(win**2*np.log2(win**2))
        ret.append(ent)
    return np.array(ret)

def katz_fd(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],sig: pcg_signal) -> npt.NDArray[np.float64]:
    """Calculate Katz fractal dimension

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        sig (pcg_signal): input signal

    Returns:
        np.ndarray: Katz fractal dimension estimates
    """
    start, end = _check_start_end(start,end)
    ret = []
    for s,e in zip(start,end):
        win = sig.data[s:e]
        ind = np.arange(len(win)-1)
        A = np.stack((ind,win[:-1]))
        B = np.stack((ind+1,win[1:]))
        dists = np.linalg.norm(B-A,axis=0)
        ind = np.arange(len(win))
        A = np.stack((ind,win))
        first = np.reshape([0,win[0]],(2,1))
        aux_d = np.linalg.norm(A-first,axis=0)
        L = np.sum(dists)
        a = np.mean(dists)
        d = np.max(aux_d)
        D = np.log10(L/a)/np.log10(d/a)
        ret.append(D)
    return np.array(ret)

def lyapunov(start: npt.NDArray[np.int_],end: npt.NDArray[np.int_],sig: pcg_signal,dim:int=4,lag:int=3) -> npt.NDArray[np.float64]:
    """Estimate Lyapunov exponent with nolds
    
    Note:
        This currently has parity problems with Matlab's lyapunovExponent function

    Args:
        start (np.ndarray): start times in samples
        end (np.ndarray): end times in samples
        sig (pcg_signal): input signal
        dim (int, optional): Lyapunov embedding dimension. Defaults to 4.
        lag (int, optional): time-lag (tau) [samples]. Defaults to 3.

    Returns:
        np.ndarray: estimated Lyapunov exponents
    """
    warnings.warn("Lyapunov exponent calculation done with nolds, which has parity problems with Matlab")
    start, end = _check_start_end(start,end)
    ret = []
    for s,e in zip(start,end):
        win = sig.data[s:e]
        ly = nolds.lyap_r(win,emb_dim=dim,lag=lag,tau=1/sig.fs) #type: ignore
        ret.append(ly)
    return np.array(ret)

class _config(TypedDict):
    calc_fun: Callable
    """funtcion for calculation"""
    name: str
    """name of feature"""
    input: Literal["env"]|Literal["raw"]
    """input signal type, `env` for envelope, `raw` for non-envelope signal"""

class feature_config(_config,total=False):
    """Type to hold feature calculation configs"""
    params: dict[str,int|float|str]
    """parameters to pass to feature calculation function"""

class feature_group:
    """Group feature calculations together for reuse
    
    Attributes:
        feature_configs (list[feature_config]): List of configurations for feature calculation.
        
    Example:
        Create a feature group:
        
        For an easier experience, use the `feature_config` type
        
        >>> import pyPCG.features as ftr
        >>> ftr_1 = {"calc_fun":ftr.time_delta,"name":"length","input":"raw"}
        >>> ftr_2 = {"calc_fun":ftr.ramp_time,"name":"onset","input":"env"}
        >>> ftr_3 = {"calc_fun":ftr.ramp_time,"name":"exit","input":"env","params":{"type":"exit"}}
        >>> my_group = ftr.feature_group(ftr_1,ftr_2,ftr_3)
        
        Run the created group with a regular signal `raw_sig` and its envelope `env_sig` for all detected S1 events (boundaries given as `s1_start`, `s1_end`):
        
        >>> my_features = my_group.run(raw_sig,env_sig,s1_start,s1_end)
    """
    def __init__(self,*configs: feature_config) -> None:
        self.feature_configs = []
        for config in configs:
            self.feature_configs.append(config)

    def run(self, raw_sig: pcg_signal, env_sig: pcg_signal, starts: npt.NDArray[np.int_], ends: npt.NDArray[np.int_]) -> dict[str,npt.NDArray[np.float64]]:
        """Run feature calculations on the input signal based on the elements of feature_configs

        Args:
            raw_sig (pcg_signal): input signal
            env_sig (pcg_signal): envelope of input signal
            starts (np.ndarray): start times in samples
            ends (np.ndarray): end times in samples

        Returns:
            dict[str,np.ndarray]: calculated features, with the names given in `feature_configs` field
        """
        ret_dict = {}
        for ftr in self.feature_configs:
            in_sig = raw_sig if ftr["input"] == "raw" else env_sig
            calc = ftr["calc_fun"](starts,ends,in_sig) if len(ftr) == 3 else ftr["calc_fun"](starts,ends,in_sig,**ftr["params"])
            ret_dict[ftr["name"]] = calc[0] if type(calc) is tuple else calc
        return ret_dict

if __name__ == '__main__':
    print("Feature calculation")


from pyPCG import pcg_signal
import scipy.signal as signal
import numpy as np
import pywt
import copy
import emd
from typing import Callable, TypedDict

def envelope(sig: pcg_signal) -> pcg_signal:
    """Calculates the envelope of the signal based on Hilbert transformation

    Args:
        sig (pcg_signal): input signal

    Returns:
        pcg_signal: envelope
    """
    ret_sig = copy.deepcopy(sig)
    ret_sig.processing_log.append("Envelope")
    ret_sig.data = np.abs(signal.hilbert(ret_sig.data)) # type: ignore
    return ret_sig

def homomorphic(sig: pcg_signal, filt_ord: int = 6, filt_cutfreq: float = 8) -> pcg_signal:
    """Calculate the homomoprphic envelope of the signal

    Args:
        sig (pcg_signal): input signal
        filt_ord (int, optional): lowpass filter order. Defaults to 6.
        filt_cutfreq (float, optional): lowpass filter cutoff frequency. Defaults to 8.

    Raises:
        ValueError: The cutoff frequency exceeds the Nyquist limit

    Returns:
        pcg_signal: homomoprhic envelope
    """
    if filt_cutfreq>sig.fs/2:
        raise ValueError("Filter cut frequency exceeds Nyquist limit")
    
    ret_sig = copy.deepcopy(sig)
    env = envelope(sig).data
    ret_sig.processing_log.append(f"Homomorphic envelope (order-{filt_ord},cut-{filt_cutfreq})")
    lp = signal.butter(filt_ord,filt_cutfreq,output='sos',fs=sig.fs,btype='lowpass')
    env[env<=0] = np.finfo(float).eps
    filt = signal.sosfiltfilt(lp,np.log(env))
    ret_sig.data = np.exp(filt)
    return ret_sig

def filter(sig: pcg_signal, filt_ord: int, filt_cutfreq: float, filt_type: str = "LP") -> pcg_signal:
    """Filters the signal based on the input parameters with a Butterworth filter design

    Args:
        sig (pcg_signal): input signal
        filt_ord (int): filter order
        filt_cutfreq (float): filter cutoff frequency
        filt_type (str, optional): filter type: `"LP"` (low-pass) or `"HP"` (high-pass). Defaults to `"LP"`.

    Raises:
        NotImplementedError: Other filter type
        ValueError: Filter cutoff exceeds Nyquist limit

    Returns:
        pcg_signal: filtered signal
    """
    longname = ""
    if filt_type == "LP":
        longname = "lowpass"
    elif filt_type == "HP":
        longname = "highpass"
    else:
        raise NotImplementedError("Only HP and LP filters are supported right now")
    if filt_cutfreq>sig.fs/2:
        raise ValueError("Filter cut frequency exceeds Nyquist limit")
    filt = signal.butter(filt_ord,filt_cutfreq,output='sos',fs=sig.fs,btype=longname)
    ret_sig = copy.deepcopy(sig)
    ret_sig.data = signal.sosfiltfilt(filt,ret_sig.data)
    ret_sig.processing_log.append(f"{filt_type} Filter (order-{filt_ord}, cut-{filt_cutfreq})")
    return ret_sig

def wt_denoise(sig: pcg_signal, th: float=0.2, wt_family: str = "coif4", wt_level: int = 5) -> pcg_signal:
    """Denoise the signal with a wavelet thresholding method

    Args:
        sig (pcg_signal): input noisy signal
        th (float, optional): threshold value given as a percentage of maximum. Defaults to 0.2.
        wt_family (str, optional): wavelet family. Defaults to "coif4".
        wt_level (int, optional): wavelet decomposition level. Defaults to 5.

    Returns:
        pcg_signal: denoised signal
    """
    ret_sig = copy.deepcopy(sig)
    th_coeffs = []
    coeffs = pywt.wavedec(ret_sig.data,wt_family,level=wt_level)
    for coeff in coeffs:
        th_coeffs.append(pywt.threshold(coeff,th*max(coeff)))
    ret_sig.data = pywt.waverec(th_coeffs,wt_family)
    ret_sig.processing_log.append(f"Wavelet denoise (family-{wt_family}, level-{wt_level}, th-{th})")
    return ret_sig

def slice_signal(sig: pcg_signal, time_len: float=60, overlap: float=0) -> list[pcg_signal]:
    """Slice long signal into a list of shorter signals

    Args:
        sig (pcg_signal): input long signal
        time_len (float, optional): desired short timelength [seconds]. Defaults to 60.
        overlap (float, optional): overlap percentage. Defaults to 0.

    Returns:
        list[pcg_signal]: list of shorter signals
    """
    time_len_s = round(time_len*sig.fs)
    step = time_len_s-round(time_len_s*overlap)
    start = 0
    acc = []
    at_end = False
    while(not at_end):
        end = start+time_len_s
        if end >= len(sig.data):
            end = len(sig.data)
            at_end = True
        sliced = pcg_signal(sig.data[start:end],sig.fs)
        acc.append(sliced)
        start += step
    return acc

def emd_denoise_sth(sig: pcg_signal) -> pcg_signal:
    """EMD based denoising with soft threshold method.
    
    Based on: Boudraa, Abdel-O & Cexus, Jean-Christophe & Saidi, Zazia. (2005). EMD-Based Signal Noise Reduction. Signal Processing. 1.
    
    Note: Tau parameter calculation modified from the original

    Args:
        sig (pcg_signal): input signal

    Returns:
        pcg_signal: denoised signal
    """
    imf = emd.sift.sift(sig.data)
    mad = np.median(np.abs(imf - np.median(imf,axis=0)),axis=0) #type: ignore
    sigma = mad/0.6745
    # MODIFIED TAU
    tau = sigma*np.sqrt(2)
    th_imf = pywt.threshold(imf,tau) #type: ignore
    ret_sig = copy.deepcopy(sig)
    ret_sig.data = np.sum(th_imf,axis=1)
    ret_sig.processing_log.append("EMD denoising (soft th)")
    return ret_sig

def emd_denoise_savgol(sig: pcg_signal, window: int=10, poly: int=3) -> pcg_signal:
    """EMD based denoising method with Savoy-Golatzky filter method.
    
    Based on: Boudraa, Abdel-O & Cexus, Jean-Christophe & Saidi, Zazia. (2005). EMD-Based Signal Noise Reduction. Signal Processing. 1.

    Args:
        sig (pcg_signal): input signal
        window (int, optional): savgol filter window size [samples]. Defaults to 10.
        poly (int, optional): savgol polynomial degree to fit. Defaults to 3.

    Returns:
        pcg_signal: denoised signal
    """
    imf = emd.sift.sift(sig.data)
    th_imf = signal.savgol_filter(imf,window,poly,mode="nearest")
    ret_sig = copy.deepcopy(sig)
    ret_sig.data = np.sum(th_imf,axis=1)
    ret_sig.processing_log.append("EMD denoising (savgol)")
    return ret_sig

def wt_denoise_sth(sig: pcg_signal, wt_family: str = "coif4", wt_level: int = 5) -> pcg_signal:
    """Denoise the signal with automatic wavelet thresholding method. Threshold is calculated automatically.
    
    Based on: D.L. Donoho, and I.M. Johnstone, Ideal spatial adaptation by wavelet shrinkage, Biometrika, vol. 81, no. 3, pp. 425-455, 1994

    Args:
        sig (pcg_signal): input noisy signal
        wt_family (str, optional): wavelet family. Defaults to "coif4".
        wt_level (int, optional): wavelet decomposition level. Defaults to 5.

    Returns:
        pcg_signal: denoised signal
    """
    ret_sig = copy.deepcopy(sig)
    th_coeffs = []
    coeffs = pywt.wavedec(ret_sig.data,wt_family,level=wt_level)
    for coeff in coeffs:
        mad = np.median(np.abs(coeff - np.median(coeff))) #type: ignore
        sigma = mad/0.6745
        # MODIFIED TAU
        tau = sigma*np.sqrt(2)
        th_coeffs.append(pywt.threshold(coeff,tau))
    ret_sig.data = pywt.waverec(th_coeffs,wt_family)
    ret_sig.processing_log.append(f"Wavelet denoise (family-{wt_family}, level-{wt_level})")
    return ret_sig

def resample(sig: pcg_signal, target_fs:int) -> pcg_signal:
    """Resample signal to target samplerate

    Args:
        sig (pcg_signal): input signal
        target_fs (int): target samplerate

    Returns:
        pcg_signal: resampled signal
    """
    ret_sig = copy.deepcopy(sig)
    ret_sig.data = signal.resample_poly(ret_sig.data,target_fs,ret_sig.fs)
    ret_sig.fs = target_fs
    ret_sig.processing_log.append(f"Resample to {target_fs} Hz")
    return ret_sig

class process_config(TypedDict):
    """Type to hold processing calculation configs"""
    step: Callable
    """function for the calculation"""
    params: dict[str,int|float|str]
    """parameters to pass to the function as keyword arguments"""

class process_pipeline:
    """Processing pipeline. One step's input is the previous step's output
    
    Attributes:
        steps (list[Callable | process_config]): List of steps as functions or function and parameters as keyword dictionary
    
    Example:
        Creating a simple pipeline
        
        >>> import pyPCG
        >>> my_pipeline = pyPCG.process_pipeline(pyPCG.zero_center, pyPCG.unit_scale)
        >>> print(my_pipeline)
        PCG processing pipeline [2 steps]
        
        Creating a pipeline with parameters:
        
        Option 1: Create a dictionary with the appropriate function with the parameters passed as keyword arguments
        
        For an easier experience, use the `process_config` type
        
        >>> import pyPCG
        >>> import pyPCG.preprocessing as preproc
        >>> step_1 = {"step":preproc.filter,"params":{"filt_ord":6,"filt_cutfreq":100,"filt_type":"LP"}}
        >>> step_2 = {"step":preproc.filter,"params":{"filt_ord":6,"filt_cutfreq":20,"filt_type":"HP"}}
        >>> my_pipeline = pyPCG.process_pipeline(step_1,step_2)
        
        Option 2: Using ``functools.partial``
        
        >>> import pyPCG
        >>> import pyPCG.preprocessing as preproc
        >>> from functools import partial
        >>> step_1 = partial(preproc.filter, filt_ord=6, filt_cutfreq=100, filt_type="LP")
        >>> step_2 = partial(preproc.filter, filt_ord=6, filt_cutfreq=20, filt_type="HP")
        >>> my_pipeline = pyPCG.process_pipeline(step_1,step_2)
        
        Use the above pipeline:
        >>> import pyPCG.io as pcg_io
        >>> data, fs = pcg_io.read_signal_file("example.wav","wav")
        >>> example = pyPCG.pcg_signal(data,fs)
        >>> processed = my_pipeline.run(example)
    """
    def __init__(self, *configs: Callable|process_config) -> None:
        """Create processing pipeline object"""
        self.steps = []
        for k in configs:
            self.steps.append(k)
            
    def __repr__(self) -> str:
        return f"PCG processing pipeline [{len(self.steps)} steps]"
    
    def run(self, input: pcg_signal) -> pcg_signal:
        """Run the processing pipeline

        Args:
            input (pcg_signal): input signal

        Returns:
            pcg_signal: processed signal
        """
        out = input
        for step in self.steps:
            if type(step) is process_config or type(step) is dict:
                out = step["step"](out,**step["params"])
            else:
                out = step(out)
        return out

if __name__ == '__main__':
    print("Preprocessing functions")